/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** cairo-gl-test
**
** main.c - pulling it all together
**
** Copyright 2014 Mirco Müller
**
** Authors:
**    Mirco "MacSlow" Mueller <macslow@gmail.com>
**
** This program is free software: you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 3, as published
** by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranties of
** MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
** PURPOSE.  See the GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program.  If not, see <http://www.gnu.org/licenses/>.
**
*******************************************************************************/

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <X11/keysymdef.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <cairo.h>
#include <cairo-gl.h>
#include <glib.h>
#include <pango/pangocairo.h>
#include <librsvg/rsvg.h>

#define WIN_WIDTH  900
#define WIN_HEIGHT 900

int fps = 0;
double frameTime = 0.0;
int frames = 0;

void
draw_checker_board (cairo_t* cr, int cols, int rows, int size)
{
	double x = 0.0f;
	double y = 0.0f;
	double width = 0.0f;
	double height = 0.0f;
	int col = 0;
	int row = 0;
	int toggle = 1;

	/* sanity-checks */
	if (cols == 0)
		cols = 1;

	if (rows == 0)
		rows = 1;

	/* determine tile-size */
	width = size / (double) cols;
	height = size / (double) rows;

	/* draw the checker-board */
	for (col = 0; col < cols; col++)
	{
		for (row = 0; row < rows; row++)
		{
			if (toggle)
				cairo_set_source_rgb (cr, 0.6, 0.6, 0.6);
			else
				cairo_set_source_rgb (cr, 0.7, 0.7, 0.7);

			cairo_rectangle (cr, x, y, width, height);
			cairo_fill (cr);

			toggle = toggle ? 0 : 1;
			x += width;
		}
		toggle = toggle ? 0 : 1;
		x = 0.0;
		y += height;
	}
}

void
demo_zini (cairo_t* cr, double size, unsigned long ms)
{
	double len = size / 25.0f;
	double y = 0.0;
	int	i = 0;
	int max = 75;

	cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_set_line_width (cr, len);

	for (i = 0; i < max; i++)
	{
		cairo_save (cr);
		cairo_translate (cr, size / 2.0, size / 2.0);
		cairo_rotate (cr, M_PI / 180.0f * (ms + 10.0f * i) * 0.36f);
		y = size*0.33f + size*0.0825f * sinf ((ms + 10.0f * i) / 1000 * 10 * M_PI);
		cairo_translate (cr, 0.0f, y);
		cairo_rotate (cr, M_PI / 180.0f * 6.0f * i);
		cairo_set_source_rgba (cr,
			1.0,
			(double) i / (double) max,
			1.0 - ((double) i / (double) max),
			i * 0.01f);
		cairo_move_to (cr, -len, 0.0f);
		cairo_line_to (cr, len, 0.0f);
		cairo_stroke (cr);
		cairo_restore (cr);
	}
}

void
calc_fps ()
{
	struct timespec tp;
	int result = 0;
	static long last = 0;
	static double sec = 0;
	static double diff = 0;

	result = clock_gettime (CLOCK_MONOTONIC, &tp);
	if (result == 0)
	{
		if (last > tp.tv_nsec)
			diff = (double) tp.tv_nsec;
		else
			diff = (double) tp.tv_nsec - (double) last;

		last = tp.tv_nsec;

		if (sec > 1.0)
		{
			frameTime = 1.0 / (double) fps;
			fps = frames;
			frames = 0;
			sec = 0.0;
		}

		sec += diff / 1000000000.0;
	}
}

PangoLayout*
setup_text_rendering (cairo_t* cr)
{
	PangoFontDescription* desc = NULL;
	PangoLayout* layout = NULL;

	/* setup a new pango-layout based on the source-context */
	layout = pango_cairo_create_layout (cr);
	if (!layout)
	{
		printf ("Could not create pango-layout!\n");
		return NULL;
	}

	/* get a new pango-description */
	desc = pango_font_description_from_string ("Ubuntu Light Italic 28");
	if (!desc)
	{
		printf ("Could not create pango-font-description!\n");
		g_object_unref (layout);
		return NULL;
	}

	pango_layout_set_font_description (layout, desc);
	pango_font_description_free (desc);

	return layout;
}

void
render_fps (cairo_t* cr, PangoLayout* layout, int frames, double frameTime)
{
	GString* textString = NULL;

	/* get buffer to hold string */
	textString = g_string_new (NULL);
	if (!textString)
	{
		printf ("Could not create string-buffer!\n");
		return;
	}

	/* print and layout string */
	g_string_printf (textString,
			 "<big>%2d</big><sub><small>fps</small></sub><big> - "
			 "%.4f</big><sub><small>frame-time</small></sub>",
			 frames, frameTime);
	pango_layout_set_markup (layout, textString->str, -1);

	g_string_free (textString, TRUE);

	/* draw the string */
	cairo_set_source_rgb (cr, 1.0f, 1.0f, 1.0f);
	cairo_move_to (cr, 5.0f, 5.0f);
	pango_cairo_layout_path (cr, layout);
	cairo_fill (cr);
}

void
draw_logo (cairo_t* cr, RsvgHandle* svg, double scale, double x, double y)
{
	cairo_save (cr);
	cairo_scale (cr, scale, scale);
	cairo_translate (cr, x, y);
	rsvg_handle_render_cairo (svg, cr);
	cairo_restore (cr);
}

static void
draw (cairo_t* cr,
	  PangoLayout* layout,
	  int width,
	  int height,
	  RsvgHandle* svg,
	  int svgWidth,
	  int svgHeight)
{
	static double i = 0.0;
	static double angle = 0.0;
	static int counter = 0;

	cairo_save (cr);
	cairo_scale (cr, 1.0, 1.0);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_set_source_rgb (cr, 1.0, 0.5, 0.25);
	cairo_paint (cr);
	draw_checker_board (cr, 10, 10, (double) width);
	//counter = (counter + 16) % 1000;
	counter = (counter + 4) % 1000;
	demo_zini (cr, (double) width, counter);
	cairo_restore (cr);

	render_fps (cr, layout, fps, frameTime);

	draw_logo (cr, svg, 1.0/3.0, 20.0*3, height*3 - svgHeight - 20.0*3);

	//cairo_surface_write_to_png (cairo_get_target (cr), "/tmp/surf.png");

	cairo_gl_surface_swapbuffers (cairo_get_target (cr));

	i += 1.1;
	angle += 1.0f;
	frames++;

	return;
}

static Window
create_window (Display* dpy, int width, int height)
{
	int aiAttribs[] = {
		GLX_RGBA,
		GLX_RED_SIZE, 1,
		GLX_GREEN_SIZE, 1,
		GLX_BLUE_SIZE, 1,
		GLX_ALPHA_SIZE, 1,
		GLX_DOUBLEBUFFER,
		None
	};

	int                  iScreen;
	XSetWindowAttributes winAttribs;
	unsigned long        ulMask;
	Window               rootWindow;
	Window               window;
	XVisualInfo*         visinfo;
	GLXContext           ctx;

	iScreen = DefaultScreen (dpy);
	rootWindow = RootWindow (dpy, iScreen);

	visinfo = glXChooseVisual (dpy, iScreen, aiAttribs);
	if (!visinfo)
	{
		printf ("Error: couldn't get an RGB, Double-buffered visual\n");
		exit (1);
	}

	/* window attributes */
	winAttribs.background_pixel = 0;
	winAttribs.border_pixel = 0;
	winAttribs.colormap = XCreateColormap (dpy,
		rootWindow,
		visinfo->visual,
		AllocNone);
	winAttribs.event_mask = StructureNotifyMask | KeyPressMask;
	ulMask = CWBackPixel |
		CWBorderPixel |
		CWColormap |
		CWEventMask;

	window = XCreateWindow (dpy,
		rootWindow,
		0,
		0,
		width,
		height,
		0,
		visinfo->depth,
		InputOutput,
		visinfo->visual,
		ulMask,
		&winAttribs);

	ctx = glXCreateContext (dpy, visinfo, NULL, False);
	if (!ctx)
	{
		printf ("Error: glXCreateContext failed\n");
		exit (1);
	}

	return window;
}

static void
event_loop (Display* dpy, Window win, cairo_t* cr)
{
	XEvent       event;
	Bool         keepGoing = True;
	Bool         gotEvent  = False;
	int          width  = 0;
	int          height = 0;
	int          newWidth  = 0;
	int          newHeight = 0;
	PangoLayout* layout = NULL;
    RsvgHandle*  svg = NULL;
	RsvgDimensionData dims;

	layout = setup_text_rendering (cr);
	svg = rsvg_handle_new_from_file ("cairo-logo.svg", NULL);
	rsvg_handle_get_dimensions (svg, &dims);

	while (keepGoing)
	{
		gotEvent = XCheckWindowEvent (dpy,
					       win,
					       KeyPressMask |
					       StructureNotifyMask,
					       &event);

		if (gotEvent)
		{
			switch (event.type)
			{
				case KeyPress :
					if (XLookupKeysym (&event.xkey, 0) == XK_Escape)
						keepGoing = False;
				break;

				case ConfigureNotify :
					newWidth  = event.xconfigure.width;
					newHeight = event.xconfigure.height;
					if (newWidth != width || newHeight != height)
					{
						width  = newWidth;
						height = newHeight;
					}
				break;
			}
		}
		else
		{
			draw (cr, layout, width, height, svg, dims.width, dims.height);
			calc_fps ();
			usleep (100);
		}
	}

	g_object_unref (layout);
}

struct closure
{
	Display *dpy;
	GLXContext ctx;
};

static void
cleanup (void *data)
{
	struct closure *arg = data;

	glXDestroyContext (arg->dpy, arg->ctx);
	XCloseDisplay (arg->dpy);

	free (arg);
}

static cairo_surface_t *
create_source_surface (Display* dpy, Window win, int width, int height)
{
    int attribs[] = {
		GLX_RGBA,
		GLX_RED_SIZE, 1,
		GLX_GREEN_SIZE, 1,
		GLX_BLUE_SIZE, 1,
		GLX_ALPHA_SIZE, 1,
		GLX_DOUBLEBUFFER,
		None
    };

    XVisualInfo*     visinfo = NULL;
    GLXContext       ctx;
    struct closure*  arg = NULL;
    cairo_device_t*  device = NULL;
    cairo_surface_t* surface = NULL;

    visinfo = glXChooseVisual (dpy, DefaultScreen (dpy), attribs);
    if (visinfo == NULL)
		return NULL;

    ctx = glXCreateContext (dpy, visinfo, NULL, True);
    free (visinfo);

    if (ctx == NULL)
		return NULL;

    arg = malloc (sizeof (struct closure));
    arg->dpy = dpy;
    arg->ctx = ctx;
    device = cairo_glx_device_create (dpy, ctx);
    if (cairo_device_set_user_data (device,
				    (cairo_user_data_key_t *) cleanup,
				    arg,
				    cleanup))
    {
		cleanup (arg);
		return NULL;
    }

    surface = cairo_gl_surface_create_for_window (device, win, width, height);

    cairo_device_destroy (device);
    glXMakeCurrent (dpy, win, ctx);

    return surface;
}

int
main (int    argc,
      char** argv)
{
	Display*         display = NULL;
	Window           win = None;
	cairo_surface_t* surf = NULL;
	cairo_t*         cr = NULL;

	dpy = XOpenDisplay (NULL);

	/* make operations synchronous */
	XSynchronize (dpy, True);

	win = create_window (dpy, WIN_WIDTH, WIN_HEIGHT);
	surf = create_source_surface (dpy, win, WIN_WIDTH, WIN_HEIGHT);
	cr = cairo_create (surf);

	XMapWindow (dpy, win);

	/* this is indeed how you change the window-title under pure X11 */
	Atom property = XInternAtom (dpy, "_NET_WM_NAME", True);
	Atom type = XInternAtom (dpy, "UTF8_STRING", True);
	XChangeProperty (dpy,
					 win,
					 property,
					 type,
					 8,
					 PropModeReplace,
					 (const unsigned char *) "cairo via OpenGL",
					 24);

	event_loop (dpy, win, cr);

	cairo_destroy (cr);
	cairo_surface_destroy (surf);

	return 0;
}
